/** Scanner. */
class AmbigiousExpLexer extends Lexer;

options {
    exportVocab = AmbigiousExp;
}

tokens {
    NUMBER<AST=NumberExp>;
    PLUS<AST=BinaryExp>;
    MINUS<AST=BinaryExp>;
    MULT<AST=BinaryExp>;
    DIV<AST=BinaryExp>;
}

WS      : (' '
          | '\t'
          | ('\n'{ newline(); })
          )+
          {$setType(Token.SKIP);}
        ;  
// PLUS    :	"+";
// MINUS	:   "-";
NUMBER  : ('0' .. '9')+ 
        ;
LITERAL_TOKENS  : "(" | ")" | ";"
        ;
PLUS    : "+"
        ;
MINUS   : "-"
        ;
MULT    : "*"
        ;
DIV     : "/"
        ;
OTHER   : ('a' .. 'z')+
    ;

/** 
 * Ambigious expression grammar, manual AST construction.
 * 
 * 
 * exp		-> exp (+ | -) | exp |
 *             exp (* | /) |
 *             NUMBER 
 *
 * Presedence rules:  
 *   + and - has lower precedence than * and /.
 * Associativity rules: 
 *   +, -, * and / are left associative.
 *
 * Problems:
 *   1. Left recursion.
 *   2. Maintain precedence (=> precedence cascade).
 *   3. Maintain associativity 
 *         (=> check that recursion direction matches associativity, if not
 *             ast nodes must be corrected by parameters to the productions). 
 *
 * Solutions:
 *   1. Rewrite the grammar.
 *   2. Rewrite the grammar (precedence cascade).
 *   3. Construct the nodes from left to right in an EBNF rule (with an 
 *      implied left to right iteration). Or, rewrite as BNF with lhs as a 
 *      parameter to the rhs recursion.
 *
 */
class AmbigiousExpParser extends Parser;

options {
    importVocab = AmbigiousExp;
}

tokens {
    PLUS="+";
    MULT="*";
}

{ 
    public static void main(String args[]) throws Exception {
        final AmbigiousExpParser parser = 
            new AmbigiousExpParser(new AmbigiousExpLexer(System.in));
        final Exp tree = parser.exp();
        System.out.println(tree.toLispString());
    }
}

// exp     : exp ("*" | "/") exp
//         ;
// Manual tree construction:
// Note: This grammar doesn't handle the precedence requirement.
exp returns[Exp e]  
{ 
    BinaryExp.Operator op; 
    e = null;
}
        : lhs:NUMBER {e = new NumberExp(Integer.valueOf(lhs.getText())); } 
          ((op=op rhs:NUMBER) 
           { e=new BinaryExp(op, e, 
                             new NumberExp(Integer.valueOf(rhs.getText()))); }
            )*
          EOF!
    ;
op returns [BinaryExp.Operator o]
{ o = null; }
        : "+" {o = Exp.Operator.PLUS; }
        | "*" {o = Exp.Operator.MULT; }
        ;



